package com.wulfson.livescore;
//
//  LSBeamGroup.java
//  LiveScore
//
//  Created by harris wulfson on 7/30/07.
//  Copyright 2007 __MyCompanyName__. All rights reserved.
//

//package

import java.awt.*;
import java.awt.geom.*;
import java.util.*;


public class LSBeamGroup
{
	LSStaff _staff;
	TreeSet _notes;
	int _stemDirection;
	float _stemEnd;
	int _beams; //experimental
	boolean _hasBeam; //could just rely on duration beams (and allow breaks)

	public LSBeamGroup()
	{
		_notes = new TreeSet(new LSNotePositionComparator());
		_beams = 1;
	}

	public void releaseAllNotes()
	{
		Iterator i;
		LSNote aNote;
		
		i=_notes.iterator();
		while(i.hasNext())
		{
			aNote=(LSNote)i.next();
			aNote.setBeamGroup(null);
			//[aNote determineStemDirection];
			//[aNote determineStemEnd];
		}
		_staff=null;
		_notes = new TreeSet(new LSNotePositionComparator());
	}

	public void beam()
	{
		this.determineStemDirection();
	}

	public void paint(Graphics g)
	{
		Graphics2D g2 = (Graphics2D)g;
		float space;
		float beamWidth;
		float leftX, rightX;
		int b;
		
		space=_staff.getSystem().getPage().getSpace();
		beamWidth = space * 0.5f; //TODO. constant.

		//since this is a TreeSet, the notes are already sorted correctly
		leftX=((LSNote)_notes.first()).getStemX();
		rightX=((LSNote)_notes.last()).getStemX();
		
		if(_hasBeam)
		{
			for(b=0; b<_beams; b++)
			{
				//TODO: this assumes space between beams = beam width.
				//The following places the stem end in the center of the beam (vertically)
				//This fixes a problem where the renderer was causing the stems to poke through
				//the beam. I have no idea which is more orthodox for music typesetting. 
				if(_stemDirection == LSNote.STEMUP)
				{
					g2.fill(new Rectangle2D.Float(leftX,
													_stemEnd + (b * beamWidth * 2) - (beamWidth / 2),
													rightX - leftX,
													beamWidth));
				}
				else
				{
					g2.fill(new Rectangle2D.Float(leftX,
										_stemEnd - (b * beamWidth * 2) - (beamWidth / 2),
										rightX - leftX,
										beamWidth));
				}
			}		
		}
		
	}

	//Calculations
	public void determineStemDirection()
	{
		Iterator i;
		LSNote aNote;
		int ups=0, downs=0;
		
		i=_notes.iterator();
		while(i.hasNext())
		{
			aNote=(LSNote)i.next();
			if(aNote.getStemDirection() == LSNote.STEMDOWN)
			{
				downs++;
			}
			else
			{
				ups++;
			}
		}
		
		if(downs > ups)
		{
			this.setStemDirection(LSNote.STEMDOWN);
		}
		else
		{
			this.setStemDirection(LSNote.STEMUP);
		}
	}
	
	public void determineStemLength()
	{
		Iterator i;
		LSNote aNote=null, winner=null;
		float min, max;
		float space;
		float beamWidth;
		
		space=_staff.getSystem().getPage().getSpace();
		beamWidth = space * 0.5f; //TODO. constant.
		
		max=0f;
		min=3000000f; //does java have a max double method?
		
		i=_notes.iterator();
		if(_stemDirection == LSNote.STEMUP)
		{
			while(i.hasNext())
			{
				aNote=(LSNote)i.next();
				if(aNote.getStemEnd() < min)
				{
					winner = aNote;
					min = aNote.getStemEnd();
				}
			}
		}
		else //find max stemend (lowest on screen)
		{
			while(i.hasNext())
			{
				aNote=(LSNote)i.next();
				if(aNote.getStemEnd() > max)
				{
					winner = aNote;
					max = aNote.getStemEnd();
				}
			}
		}
		
		if(winner == null) { winner=aNote; } //very unlikely case 
		if(_stemDirection == LSNote.STEMUP)
		{
			_stemEnd=winner.getStemEnd() - (_beams-1) * (beamWidth * 2); //TODO: this assumes space between beams = beam width. 
		}
		else
		{
			_stemEnd=winner.getStemEnd() + (_beams-1) * (beamWidth * 2);
		}
		
		this.publishStemEnds();
		
	}
	
	public void publishStemEnds()
	{
		Iterator i;
		LSNote aNote;
		
		i=_notes.iterator();
		while(i.hasNext())
		{
			aNote=(LSNote)i.next();
			aNote.setStemEnd(_stemEnd);
		}
	}


	//Accessors
	public LSStaff getStaff()
	{
		return _staff;
	}

	public void setStaff(LSStaff aStaff)
	{
		_staff = aStaff;
	}

	public TreeSet getNotes()
	{
		return _notes;
	}

	public void addNote(LSNote aNote)
	{
		_notes.add(aNote);
		aNote.setBeamGroup(this);
	}
	
	public void removeNote(LSNote aNote)
	{
		_notes.remove(aNote);
		aNote.setBeamGroup(null);
	}

	public int getStemDirection()
	{
		return _stemDirection;
	}

	public void setStemDirection(int stemDirection)
	{
		Iterator i;
		LSNote aNote;
		
		_stemDirection = stemDirection;
		i=_notes.iterator();
		while(i.hasNext())
		{
			aNote=(LSNote)i.next();
			aNote.setStemDirection(stemDirection);
			//aNote.setHasStem(true);
		}
		this.determineStemLength();
		
	}

	public float getStemEnd()
	{
		return _stemEnd;
	}

	public void setStemEnd(float aStemEnd)
	{
		_stemEnd = aStemEnd;
	}

	public boolean getHasBeam()
	{
		return _hasBeam;
	}

	public void setHasBeam(boolean hasBeam)
	{
		_hasBeam = hasBeam;
	}	
	

}
