package com.wulfson.livescore;
//
//  LSNote.java
//  LiveScore
//
//  Created by harris wulfson on 7/29/07.
//  Copyright 2007 __MyCompanyName__. All rights reserved.
//

//package

import java.awt.*;
import java.awt.geom.*;
//import java.util.*;

public class LSNote 
{
	static final int NOACCIDENTAL = 0;
	static final int NATURAL = 1;
	static final int SHARP = 2;
	static final int FLAT = 3;
	static final String flatString = "b";
	static final String naturalString = "n";
	static final String sharpString = "#";

	static final int NODYNAMIC = 0;
	static final int PIANISSIMO = 1;
	static final int PIANO = 2;
	static final int MEZZOPIANO = 3;
	static final int MEZZOFORTE = 4;
	static final int FORTE = 5;
	static final int FORTISSIMO = 6;
	//TODO: use hex encodings	
	static final String pianissimoString = "";
	static final String pianoString = "p";
	static final String mezzoPianoString = "p";
	static final String mezzoForteString = "f";
	static final String forteString = "f";
	static final String fortissimoString = "";
		
	static final int STEMUP = 1;
	static final int STEMDOWN = 0;	

	double _onset; //seconds since start time
	double _duration;
	LSStaff _staff;
	LSBeamGroup _beamGroup;
	float _xpos;
	int _steps; //middle line = 0
	int _accidental;
	int _stemDirection;
	float _durationBeamLength; //in spaces
	float _stemEnd;
	float _stemX;
	boolean _hasStem;
	boolean _hasNotehead;
	int _notehead;
	boolean _tremolo;
	int _dynamic;
	String _text; //text below the staff. lyric.
	String _instruction; //tex above the staff. technique / expression.
	//notehead, etc...
	//list of notations (dynamics, crec, etc that also could be attached to a staff directly)

	public LSNote(LSStaff staff, float xpos, int steps, int accidental)
	{
		_hasStem=true;
		_hasNotehead=true;
		_notehead=339;
		_tremolo=false;
		_dynamic=0;
		_staff = staff;
		_accidental = accidental;
		this.setXpos(xpos);
		this.setSteps(steps);
	}

	public void setPitch(int value)
	{		
		int octave;
		int map[][] =
			{
				{0,0},
				{0,SHARP},
				{1,0},
				{1,SHARP},
				{2,0},
				{3,0},
				{3,SHARP},
				{4,0},
				{4,SHARP},
				{5,0},
				{5,SHARP},
				{6,0}
			};
		//If arg is 0, then use center line. TODO make convenience method for this block
		if(value == 0)
		{
			this.setSteps(0);
			this.setAccidental(0);
			this.setHasNotehead(false);
			return;
		}
		
		//Else
		octave=((value - (value % 12)) / 12) - 1;
		this.setSteps(map[value % 12][0] + _staff.getClef().getMiddleCSteps() + ((octave-4)*7));
		this.setAccidental(map[value % 12][1]);
		
	}
	
	public void paint(Graphics g)
	{
		Graphics2D g2 = (Graphics2D)g;
		float noteYOffset;
		float space;
		float ypos;
		float bottomY; //used to place text and then dynamic
		float topY; //used to place instruction

		//ledger line vars
		int lldirection;
		int numLedgerLines;
		int lidx;
		float llypos;
		
		//note & accidental
		//GlyphVector glyph;
		String noteheadString;
		float accidentalLead;
		String dynamicString;
		float durationBeamLength;
		float nextSystemDurationBeamLength = 0;
		
		//beam
		float beamWidth;
		
		//Ledger lines, accidental, notehead, stem, tremolo, duration beam, dynamic	
		space=this.getStaff().getSystem().getPage().getSpace();
		noteYOffset=(space * 0.085f); 
		ypos=this.getYpos();

		//Draw ledger lines.
		if(Math.abs(_steps) > 5) //do we need ledger lines?
		{
			if(_steps < 0) //below the staff?
			{
				lldirection = -1;
			}
			else //or above?
			{
				lldirection = 1;
			}
			//TODO: these calculations work, but got mangled in port from obj-c. simplify?
			numLedgerLines=Math.abs((int)(_steps/2 - (2 * lldirection)));
			g2.setStroke(new BasicStroke(space * 0.1f)); //TODO: consider caching BasicStroke object (used also by stems)
			for(lidx=0;lidx < numLedgerLines; lidx++)
			{
				llypos=_staff.getYpos() + ((lidx + 3) * space) * lldirection * -1;
				g2.draw(new Line2D.Float(_xpos - (space * 0.4f), llypos, _xpos + (space * 1.6f), llypos));
			}
		}


		//Draw accidental. TODO: break into seperate method		
		g2.setFont(new Font("Maestro", Font.PLAIN, (int)(space * 4.0f))); //TODO
		
		accidentalLead = space * 1.2f;
		switch(_accidental)
		{
			case NOACCIDENTAL:
				break;
			
			//TODO: this constant 1.1 should be an instance variable
			case NATURAL:
				g2.drawString(naturalString, this.getXpos() - accidentalLead, ypos + noteYOffset);
				break;
			case SHARP:
				g2.drawString(sharpString, this.getXpos() - accidentalLead, ypos + noteYOffset);
				break;
			case FLAT:
				g2.drawString(flatString, this.getXpos() - accidentalLead, ypos + noteYOffset);
				break;
		}
		
		//Draw notehead
		//GlyphVector looked promising, but it too is only Float precision. 
		//glyph=_staff.getSystem().getPage().getFont().createGlyphVector(g2.getFontRenderContext(), 339);
		//java doesn't appear to have double precision for positioning text??????
		if(_hasNotehead)
		{
			noteheadString=new Character((char)_notehead).toString();
			g2.drawString(noteheadString, this.getXpos(), ypos + noteYOffset);
		}
		
		//Draw stem
		if(_hasStem)
		{
			g2.setStroke(new BasicStroke(space * 0.1f));
			if(_stemDirection == LSNote.STEMUP)
			{
				g2.draw(new Line2D.Float(_stemX, ypos-(space/6), _stemX, _stemEnd));
			}
			else
			{
				g2.draw(new Line2D.Float(_stemX, ypos+(space/6), _stemX, _stemEnd));
			}
		}
		
		//Draw duration beam
		//TODO: should have more precise placement of right X. doesn't match conductor.
		//TODO: should have more precise placement of right X. doesn't match conductor.
		if(_durationBeamLength > 0)
		{
			durationBeamLength = _durationBeamLength;
			//Cut if off at the right margin.
			if(_durationBeamLength + _stemX > _staff.getSystem().getRightX())
			{
				durationBeamLength = _staff.getSystem().getRightX() - _stemX;
				nextSystemDurationBeamLength = _durationBeamLength - durationBeamLength + _staff.getSystem().getPage().getRightXInset();
			}
			
			beamWidth = space * 0.5f; //TODO: make 0.5 (beam width) variable
			//TODO: aren't the if and else blocks equivalent?
			if(_stemDirection == LSNote.STEMUP)
			{
				g2.fill(new Rectangle2D.Float(_stemX, _stemEnd - (beamWidth/2), durationBeamLength, beamWidth));
			}
			else
			{
				g2.fill(new Rectangle2D.Float(_stemX, _stemEnd - (beamWidth/2), durationBeamLength, beamWidth));
			}

			//If we had to truncate the beam, draw the rest on the next system
			if(nextSystemDurationBeamLength > 0)
			{	//TODO: something is not exact about the end location here. it goes slightly too long
				g2.fill(new Rectangle2D.Float(_staff.getSystem().getPage().getLeftMargin() + _staff.getSystem().getPage().getLeftXInset(), ((LSStaff)_staff.getSystem().getNextSystem().getStaves().get(
							this.getStaff().getSystem().getStaves().indexOf(this.getStaff()) //part index
							)).getYpos() + (_stemEnd - _staff.getYpos()) - (beamWidth/2), nextSystemDurationBeamLength, beamWidth));
			}
			
		}
		
		//Draw instruction
		topY = this.getTopY() - (2*space);
		if(_instruction != null)
		{
			g2.setFont(new Font("Helvetica", Font.ITALIC, (int)(2 * space + 0.5))); //TODO use frame to get font			
			g2.drawString(_instruction, this.getXpos() - (3*space), topY);
		}
		
		//Draw text		
		bottomY = this.getBottomY() + (4*space); //position of next lowest thing
		if(_text != null && ! _text.equals(""))
		{
			g2.setFont(new Font("Helvetica", Font.BOLD, (int)(2.3 * space + 0.5))); //TODO use frame to get font			
			g2.drawString(_text, this.getXpos() - (1*space), bottomY);
			bottomY = bottomY + (3*space); //advance position of next lowest thing
		}
		
		//Draw dynamic. TODO: break into separate method
		if(_dynamic != NODYNAMIC)
		{
			dynamicString = "";
			switch(_dynamic)
			{				
				//TODO: this constant 1.1 should be an instance variable
				case PIANISSIMO:
					dynamicString = pianissimoString;
					break;
				case PIANO:
					dynamicString = pianoString;
					break;
				case MEZZOPIANO:
					dynamicString = mezzoPianoString;
					break;
				case MEZZOFORTE:
					dynamicString = mezzoForteString;
					break;
				case FORTE:
					dynamicString = forteString;
					break;
				case FORTISSIMO:
					dynamicString = fortissimoString;
					break;
			}
			//TODO this re-uses acidentalLead. should be separate value.
			g2.setFont(new Font("Maestro", Font.PLAIN, (int)(space * 4.0f))); //TODO
			g2.drawString(dynamicString, this.getXpos() - accidentalLead, bottomY);
		}
		
	}

	//Calculate methods
	public void determineStemDirection()
	{
		if(_steps < 0)
		{
			this.setStemDirection(LSNote.STEMUP);
		}
		else
		{
			this.setStemDirection(LSNote.STEMDOWN);
		}
		
	}

	public void determineStemEnd()
	{
		float space;
		space=this.getStaff().getSystem().getPage().getSpace();


		//TODO: this calculation would be easier if the constants were 1 and -1 (thus overloaded)
		if(_stemDirection == LSNote.STEMUP)
		{
			if(_steps < -7)
				_stemEnd=_staff.getYpos();
			else
				_stemEnd=this.getYpos() - (space/6) - (space*3.334f);
		}
		else
		{
			if(_steps > 7)
				_stemEnd=_staff.getYpos();
			else
				_stemEnd=this.getYpos() + (space/6) + (space*3.344f);
		}
	}

	public void determineStemX()
	{
		float space;
		space=this.getStaff().getSystem().getPage().getSpace();

		//TODO: there is a problem with using double precision for the stems,
		//but only float for the noteheads. sometimes they are out of alignment.
		//probably need to convert stems to float (lowest common precision).
		if(_stemDirection == LSNote.STEMUP)
		{
			_stemX = _xpos + (space * 1.1f);
		}
		else
		{
			_stemX = _xpos - (space * 0.1f);
		}
	}

	public float getTopY()
	{
		float space;
		float topY;
		float check;
		
		space=this.getStaff().getSystem().getPage().getSpace();
		
		topY = this.getYpos(); //start with position of notehead

		check = _staff.getYpos() - (space*2); //check if bottom of staff is lower
		if(check < topY)
		{
			topY = check;
		}
		
		check = this.getStemEnd(); // check if stem end is lower
		if(check < topY)
		{
			topY = check;
		}
		return topY;
	}
	public float getBottomY()
	{
		float space;
		float bottomY;
		float check;
		
		space=this.getStaff().getSystem().getPage().getSpace();
		
		bottomY = this.getYpos(); //start with position of notehead

		check = _staff.getYpos() + (space*2); //check if bottom of staff is lower
		if(check > bottomY)
		{
			bottomY = check;
		}
		
		check = this.getStemEnd(); // check if stem end is lower
		if(check > bottomY)
		{
			bottomY = check;
		}
		return bottomY;
	}

	//Accessors
	
	public double getOnset()
	{
		return _onset;
	}

	public void setOnset(double onset)
	{
		_onset = onset;
	}

	public double getDuration()
	{
		return _duration;
	}

	public void setDuration(double duration)
	{
		_duration = duration;
	}

	public LSStaff getStaff()
	{
		return _staff;
	}

	public void setStaff(LSStaff aStaff)
	{
		_staff = aStaff;
	}

	public LSBeamGroup getBeamGroup()
	{
		return _beamGroup;
	}

	public void setBeamGroup(LSBeamGroup aBeamGroup)
	{
		_beamGroup = aBeamGroup;
	}

	public float getXpos()
	{
		return _xpos;
	}

	public void setXpos(float anXpos)
	{
		_xpos = anXpos;
		this.determineStemX();
	}

	public int getSteps()
	{
		return _steps;
	}

	public void setSteps(int aSteps)
	{
		_steps = aSteps;
		if(_beamGroup == null)
		{
			this.determineStemDirection();
		}
	}

	public int getAccidental()
	{
		return _accidental;
	}

	public void setAccidental(int anAccidental)
	{
		_accidental = anAccidental;
	}

	public int getStemDirection()
	{
		return _stemDirection;
	}

	public void setStemDirection(int aStemDirection)
	{
		_stemDirection = aStemDirection;
		this.determineStemX();
		this.determineStemEnd();		
	}

	public float getDurationBeamLength()
	{
		return _durationBeamLength;
	}

	public void setDurationBeamLength(float aDurationBeamLength)
	{
		_durationBeamLength = aDurationBeamLength;
	}

	public float getStemEnd()
	{
		return _stemEnd;
	}

	public void setStemEnd(float aStemEnd)
	{
		_stemEnd = aStemEnd;
	}

	public float getStemX()
	{
		return _stemX;
	}

	public void setStemX(float aStemX)
	{
		_stemX = aStemX;
	}

	public boolean getHasStem()
	{
		return _hasStem;
	}

	public void setHasStem(boolean anHasStem)
	{
		_hasStem = anHasStem;
	}

	public boolean getHasNotehead()
	{
		return _hasNotehead;
	}

	public void setHasNotehead(boolean hasNotehead)
	{
		_hasNotehead = hasNotehead;
	}

	public int getNotehead()
	{
		return _notehead;
	}

	public void setNotehead(int aNotehead)
	{
		_notehead = aNotehead;
	}

	public boolean getTremolo()
	{
		return _tremolo;
	}

	public void setTremolo(boolean aTremolo)
	{
		_tremolo = aTremolo;
	}

	public int getDynamic()
	{
		return _dynamic;
	}

	public void setDynamic(int aDynamic)
	{
		_dynamic = aDynamic;
	}
	public String getText()
	{
		return _text;
	}
	public void setText(String text)
	{
		_text = text;
	}
	public String getInstruction()
	{
		return _instruction;
	}
	public void setInstruction(String instruction)
	{
		_instruction = instruction;
	}

	//calculated
	public float getYpos()
	{
		return _staff.getYpos() - (_steps * (_staff.getSystem().getPage().getSpace() / 2));
	}

}
